use aya_ebpf::{
    bindings::xdp_action,
    macros::map,
    programs::XdpContext,
    maps::PerfEventArray,
    maps::PerCpuArray,
};
use aya_log_ebpf::info;
use net_logger_common::Event;
use core::mem;
use network_types::{
    eth::{EthHdr, EtherType},
    ip::{IpProto, Ipv4Hdr, Ipv6Hdr, in6_addr},
    tcp::TcpHdr,
    udp::UdpHdr,
    icmp::IcmpHdr,
};

#[map]
pub static mut SCRATCH: PerCpuArray<Event> = PerCpuArray::with_max_entries(1, 0); // per-cpu

#[map]
pub static mut EVENTS: PerfEventArray<Event> = PerfEventArray::with_max_entries(0, 0);

#[inline(always)] //
fn ptr_at<T>(ctx: &XdpContext, offset: usize) -> Result<*const T, ()> {
    let start = ctx.data();
    let end = ctx.data_end();
    let len = mem::size_of::<T>();

    if start + offset + len > end {
        return Err(());
    }

    Ok((start + offset) as *const T)
}

pub fn trafic_router(ctx: XdpContext) -> Result<u32, ()> {
    let ethhdr: *const EthHdr = ptr_at(&ctx, 0)?; //
    match unsafe { (*ethhdr).ether_type } {
        EtherType::Ipv4 => process_v4(ctx),
        EtherType::Ipv6 => process_v6(ctx),
        _ => Ok(xdp_action::XDP_PASS),
    }
}

pub fn get_ports(ctx: &XdpContext, protocol: IpProto, ip_hdr_len: usize) -> Result<(u16, u16),()> {
    let source_port;
    let dest_port;
    match protocol {
        IpProto::Tcp => {
            let tcphdr: *const TcpHdr =
                ptr_at(ctx, EthHdr::LEN + ip_hdr_len)?;
            source_port = u16::from_be(unsafe { (*tcphdr).source });
            dest_port = u16::from_be(unsafe { (*tcphdr).dest });
        }
        IpProto::Udp => {
            let udphdr: *const UdpHdr =
                ptr_at(ctx, EthHdr::LEN + ip_hdr_len)?;
            source_port = u16::from_be(unsafe { (*udphdr).source });
            dest_port = u16::from_be(unsafe { (*udphdr).dest });
        }
        _ => {
            source_port = 0;
            dest_port = 0;
        },
    };
    Ok((source_port, dest_port))
}

pub fn process_v6(ctx: XdpContext) -> Result<u32, ()> {
    let ipv6hdr: *const Ipv6Hdr = ptr_at(&ctx, EthHdr::LEN)?;
    let source_addr = unsafe { (*ipv6hdr).src_addr() };
    let dest_addr = unsafe { (*ipv6hdr).dst_addr() };
    let protocol = unsafe { (*ipv6hdr).next_hdr };
    let pack_len = unsafe { (*ipv6hdr).payload_len };
    let ports = get_ports(&ctx, protocol, Ipv6Hdr::LEN)?;

    let mut event = unsafe { *SCRATCH.get_ptr_mut(0).ok_or(())? };
    event.ipv = EtherType::Ipv6;
    event.source_port = ports.0;
    event.source_ipv6 = source_addr;
    event.dest_port = ports.1;
    event.dest_ipv6 = dest_addr;
    event.proto = protocol;
    event.len = pack_len;
    unsafe { EVENTS.output(&ctx, &mut event, 0); }

    Ok(xdp_action::XDP_PASS)
}

pub fn process_v4(ctx: XdpContext) -> Result<u32, ()> {

    let ipv4hdr: *const Ipv4Hdr = ptr_at(&ctx, EthHdr::LEN)?;
    let source_addr = u32::from_be(unsafe { (*ipv4hdr).src_addr });
    let dest_addr = u32::from_be(unsafe { (*ipv4hdr).dst_addr });
    let protocol = unsafe { (*ipv4hdr).proto };
    let pack_len = unsafe { (*ipv4hdr).tot_len };
    let ports = get_ports(&ctx, protocol, Ipv4Hdr::LEN)?;

    let mut event = unsafe { *SCRATCH.get_ptr_mut(0).ok_or(())? };

    event.ipv = EtherType::Ipv4;
    event.source_port = ports.0;
    event.source_ipv4 = source_addr;
    event.dest_port = ports.1;
    event.dest_ipv4 = dest_addr;
    event.proto = protocol;
    event.len = pack_len;
    unsafe { EVENTS.output(&ctx, &mut event, 0); }

    Ok(xdp_action::XDP_PASS)
}
