use anyhow::Context;
use aya::maps::AsyncPerfEventArray;
use aya::programs::{Xdp, XdpFlags};
use aya::util::online_cpus;
use aya::{include_bytes_aligned, Bpf};
use aya_log::BpfLogger;
use aya_log::Formatter;
use bytes::BytesMut;
use clap::Parser;
use log::{debug, info, warn};
use net_logger_common::Event;
use network_types::eth::EtherType;
use tokio::signal;
mod clasificator;
mod elk;
mod utils;

#[derive(Debug, Parser)]
struct Opt {
    #[clap(short, long, default_value = "lo")]
    iface: String,
}

#[tokio::main]
async fn main() -> Result<(), anyhow::Error> {
    env_logger::init();

    // Bump the memlock rlimit. This is needed for older kernels that don't use the
    // new memcg based accounting, see https://lwn.net/Articles/837122/
    let rlim = libc::rlimit {
        rlim_cur: libc::RLIM_INFINITY,
        rlim_max: libc::RLIM_INFINITY,
    };
    let ret = unsafe { libc::setrlimit(libc::RLIMIT_MEMLOCK, &rlim) };
    if ret != 0 {
        debug!("remove limit on locked memory failed, ret is: {}", ret);
    }

    if let Err(e) = load_bpf().await {
        eprintln!("error: {:#}", e);
    }

    Ok(())
}

async fn load_bpf() -> Result<(), aya::BpfError> {
    // This will include your eBPF object file as raw bytes at compile-time and load it at
    // runtime. This approach is recommended for most real-world use cases. If you would
    // like to specify the eBPF program at runtime rather than at compile-time, you can
    // reach for `Bpf::load_file` instead.
    #[cfg(debug_assertions)]
    let mut bpf = Bpf::load(include_bytes_aligned!(
        "../../target/bpfel-unknown-none/debug/net-logger"
    ))?;
    #[cfg(not(debug_assertions))]
    let mut bpf = Bpf::load(include_bytes_aligned!(
        "../../target/bpfel-unknown-none/release/net-logger"
    ))?;
    if let Err(e) = BpfLogger::init(&mut bpf) {
        // This can happen if you remove all log statements from your eBPF program.
        warn!("failed to initialize eBPF logger: {}", e);
    }

    proces_events_xdp(bpf).await
}

async fn proces_events_xdp(mut bpf: Bpf) -> Result<(), aya::BpfError> {
    let opt = Opt::parse();

    // This will include your eBPF object file as raw bytes at compile-time and load it at
    // runtime. This approach is recommended for most real-world use cases. If you would
    // like to specify the eBPF program at runtime rather than at compile-time, you can
    // reach for `Bpf::load_file` instead.

    //let program : &mut SockOps = bpf.program_mut("get_package_data").unwrap().try_into().unwrap();
    //program.load().unwrap();
    let program: &mut Xdp = bpf.program_mut("net_logger").unwrap().try_into().unwrap();
    program.load().unwrap();
    program.attach(&opt.iface, XdpFlags::default())
        .context("failed to attach the XDP program with default flags - try changing XdpFlags::default() to XdpFlags::SKB_MODE").unwrap();
    let cpus = online_cpus().unwrap();
    let num_cpus = cpus.len();
    //let events_raw = bpf.map_mut("EVENTS");
    let mut events: AsyncPerfEventArray<_> = bpf
        .take_map("EVENTS")
        .context("failed to map QUERY_RING")
        .unwrap()
        .try_into()
        .unwrap();
    for cpu in cpus {
        let mut buf = events.open(cpu, None).unwrap();

        tokio::task::spawn(async move {
            let mut buffers = (0..num_cpus)
                .map(|_| BytesMut::with_capacity(10240))
                .collect::<Vec<_>>();
            let mut con = clasificator::store::net_stats_storage::new();
            loop {
                let events = buf.read_events(&mut buffers).await.unwrap();
                for i in 0..events.read {
                    // get timestamp
                    //let now = Local::now();

                    // read the event
                    let buf = &mut buffers[i];
                    let ptr = buf.as_ptr() as *const Event;
                    let data = unsafe { ptr.read_unaligned() };
                    // parse out the data
                    match data.ipv {
                        EtherType::Ipv4 => {
                            println!("source ip:{}, source port:{}, dest ip:{}, dest port:{}, proto:{}, size: {}",
                                aya_log::Ipv4Formatter::format(data.source_ipv4),
                                data.source_port,
                                aya_log::Ipv4Formatter::format(data.dest_ipv4),
                                data.dest_port,
                                utils::ip::print_proto(data.proto),
                                data.len);
                            con.addv4(
                                data.source_ipv4,
                                data.dest_ipv4,
                                data.source_port,
                                data.dest_port,
                                data.proto,
                                data.len,
                            )
                            .await;
                            //println!("package size:{}, ips:{}", con.len, con.ip4_list.get(con.ip4_list.len()-1).unwrap().proto_len.get(0).unwrap_or(&(0,0)).1);
                        }
                        EtherType::Ipv6 => {
                            println!("source ip:{}, source port:{}, dest ip:{}, dest port:{}, proto:{}, size: {}",
                                aya_log::Ipv6Formatter::format(data.source_ipv6),
                                data.source_port,
                                aya_log::Ipv6Formatter::format(data.dest_ipv6),
                                data.dest_port,
                                utils::ip::print_proto(data.proto),
                                data.len);
                            con.addv6(
                                data.source_ipv6,
                                data.dest_ipv6,
                                data.source_port,
                                data.dest_port,
                                data.proto,
                                data.len,
                            )
                            .await;
                        }
                        _ => println!("result not coverd"),
                    }
                }
            }
        });
    }

    info!("Waiting for Ctrl-C...");
    signal::ctrl_c().await.unwrap();
    info!("Exiting...");

    Ok(())
}
